import QtQuick 1.1
import Box2D 1.0
import Qt.labs.particles 1.0
import "data.js" as B
import "game.js" as G

Item {
    id: gameArea
    anchors.fill: parent
    focus: true
    visible: false;

    property bool debug: false

    // Is the game area active and game is playing ?
    property bool active: false;

    // Is game active but in paused state, set by user or outside event (unfocused, minimized) ?
    property bool paused: state=='paused';

    // The cycle the player controls
    property variant cycle: null;
    property bool finished: false; // is game done, game over or in goal
    property bool dead: false;
    property bool displayScore: true
    property bool useZoom: _useScaling && _isFast
    property bool useBackground: _useBackground;

    // The player
    // property variant player: null;

    // The level to load
    property string level: ""
    property string levelDescription: ""
    property int levelHighScore: 0;

    signal gameEnded();
    signal gameError();
    signal gameLoaded();
    signal gamePaused(bool paused);

    state: "starting"

    property string endState: ""

    states: [
        State { name: "starting"; when: !active; },
        State { name: "playing"; when: active },
        State { name: "paused"; when: !finished; },
        State { name: "stopped";  },
        State { name: "complete"; when: finished && !dead; },
        State { name: "dead"; when: dead; },
        State { name: "noWater"; },
        State { name: "noEnergy"; }
    ]

    function doGameEndActions(msg) {
        finished=true;
        world.zoomClose();
        gameEnd.setScore(0, 0, false);
        gameEnd.setMessage(msg);
        gameEnd.show();
    }

    transitions: [
        Transition {
            from: "starting"
            to: "playing"
            ScriptAction { script: { active=true; } }
        },
        Transition {
            from: "playing"
            to: "paused"
            ScriptAction {
                script: { gamePausedDialog.show(); }
            }
        },
        Transition {
            from: "paused"
            to: "playing"
            ScriptAction {
                script: { gamePausedDialog.hide(); }
            }
        },
        Transition {
            from: "paused"
            to: "noWater"
            ScriptAction {
                script: {
                    gamePausedDialog.hide();
                    doGameEndActions(qsTr("You ran out of water!"));
                }
            }
        },
        Transition {
            from: "paused"
            to: "noEnergy"
            ScriptAction {
                script: {
                    gamePausedDialog.hide();
                    doGameEndActions(qsTr("You ran out of water!"));
                }
            }
        },
        Transition {
            from: "playing"
            to: "noWater"
            ScriptAction {
                script: {                    
                    doGameEndActions(qsTr("You ran out of water!"));
                }
            }
        },
        Transition {
            from: "playing"
            to: "noEnergy"
            ScriptAction {
                script: {                    
                    doGameEndActions(qsTr("You ran out of energy!"));
                }
            }
        },
        Transition {
            from: "playing"
            to: "dead"
            ScriptAction {
                script: {                    
                    doGameEndActions(qsTr("You had an accident!"));
                }
            }
        },
        Transition {
            from: "paused"
            to: "dead"
            ScriptAction {
                script: {
                    doGameEndActions(qsTr("You had an accident!"));
                }
            }
        },
        Transition {
            from: "playing"
            to: "complete"
            ScriptAction {
                script: {
                    var s;

                    finished=true;
                    world.zoomClose();

                    s=player.getScore()+player.getScoreBonus();
                    gameEnd.setScore(s, player.getDroplets(), s>levelHighScore ? true : false);
                    gameEnd.setMessage(qsTr("Congratulations!"));
                    if (s>levelHighScore)
                        _settings.setInt(G.getScoreId(gameArea.level), s);
                    gameEnd.show();

                }
            }
        }
    ]

    Component.onCompleted: {
        B.initFoodItems();
        levelHighScore=_settings.getInt(G.getScoreId(level), 0);
        if (levelHighScore<0)
            levelHighScore=0;
        console.debug("LEVEL HIGH SCORE IS: "+levelHighScore);
        // Hidden settings from .ini
        var customSettings=_settings.getBool("customSettings", false);
        if (customSettings) {
            world.reportFps=_settings.getBool("showFps", false);
            _useOpacity=_settings.getBool("useOpacity", true);
            useZoom=_settings.getBool("useZoom", true);
            _isFast=_settings.getBool("isFast", true);
        }
    }

    function pedal()
    {
        if (!player.haveEnergy) {
            player.stopUseEnergy();
            world.zoomNormal();
            return;
        }
        var p=cycle.pedal();
        if (cycle.inAir)
            cycle.goForward();

        if (p)
            player.startUseEnergy();
        world.zoomNormalSpeed();
    }

    function stopPedal()
    {
        player.stopUseEnergy();
        stopAll();
    }

    function jump()
    {
        if (!player.haveEnergy) {
            stopAll();
            return;
        }
        if (cycle.jump())
            player.useEnergy(5);
    }

    function brake()
    {
        if (!player.haveEnergy)
            return;
        var p=cycle.brake();
        if (cycle.inAir)
            cycle.goBackward();

        player.useEnergy(p ? 2 : 1);
    }

    function stopAll()
    {
        cycle.stopPedal();
        player.stopUseEnergy();
        world.zoomNormal();
    }

    function stopBrake()
    {
        cycle.stopBrake();
    }

    function stopJump()
    {
        cycle.stopJump();
    }

    Keys.onPressed: {
        if (gameArea.state==='starting')
            active=true;

        if (paused)
            return;

        if (event.isAutoRepeat)
            return;

        if (!cycle) {
            console.debug("Cycle is not defnied!")
            return;
        }

        switch (event.key) {
        case Qt.Key_Right:
            if (!finished) pedal();
            break;
        case Qt.Key_Left:
            if (!finished) brake();
            break;
        case Qt.Key_Up:
            if (!finished) jump();
            break;
        case Qt.Key_R:
            restartGame();
            break;
        case Qt.Key_Q:
            stopGame();
            break;
        case Qt.Key_Escape:
            stopGame();
            break;
        case Qt.Key_Plus:
            world.zoomIn();
            break;
        case Qt.Key_Minus:
            world.zoomOut();
            break;
        case Qt.Key_End:
            world.zoomOutDebug();
            break;
        case Qt.Key_D:
            debug=!debug;
            break;
        case Qt.Key_S:
            displayScore=!displayScore;
            break;
        case Qt.Key_E:
            player.energy=0;
            break;
        case Qt.Key_W:
            player.water=0;
            break;
        case Qt.Key_Z:
            useZoom=!useZoom;
            break;
        case Qt.Key_F:
            world.reportFps=!world.reportFps
            break;
        case Qt.Key_B:
            useBackground=!useBackground;
            break;
        case Qt.Key_J:
            if (debug) cycle.stepDown();
            break;
        }

        event.accepted = true;
    }

    Keys.onReleased: {
        if (paused)
            return;

        if (event.isAutoRepeat)
            return;

        if (!cycle) {
            console.debug("Cycle is not defnied!")
            return;
        }

        switch (event.key) {
        case Qt.Key_Left:
            if (!finished) cycle.stopBrake();
            break;
        case Qt.Key_Right:
            if (!finished) stopPedal();
            break;
        case Qt.Key_Up:
            if (!finished) stopJump();
            break;
        }

        event.accepted = true;
    }

    function restartGame()
    {
        player.reset();
        world.zoomNormal();
        world.restartCycle();
        gameTimer.reset();
        state='starting';
    }

    function startGame()
    {
        state='playing';
    }

    function stopGame()
    {
        state='stopped';
        gameEnded();
    }

    function startGameEnd()
    {
        state=gameArea.endState;
    }

    function pauseGame()
    {
        state='paused';
    }

    Player {
        id: player;

        onNoEnergyLeft: {
            console.debug("*** Ran out of energy");
            gameArea.endState='noEnergy';
            stopPedal();
            startGameEnd();
        }
        onNoWaterLeft: {
            console.debug("*** Ran out of water");
            gameArea.endState='noWater';
            stopPedal();
            startGameEnd();
        }

        onPlayerDead: {
            console.debug("*** Player died");
            gameArea.endState='dead';
            gameStopTimer.start();
        }

        onVechicleBroke: {
            console.debug("*** Player vehicle brokedown");
            gameArea.endState='dead';
            gameStopTimer.start();
        }

        onCriticalEnergyChanged: {
            console.debug("*** Critical Energy");
            energyMeter.setCritical(criticalEnergy);
        }

        onCriticalWaterChanged: {
            console.debug("*** Critical water");
            waterMeter.setCritical(criticalWater);
        }
    }

    Timer {
        id: waterBonusTimer
        interval: 5000;
        onTriggered: {
            // XXX: Drop a water item?
        }
    }

    Timer {
        id: gameStopTimer
        running: false;
        interval: 3500;
        onTriggered: startGameEnd();
    }

    Background {
        id: bg
        x: getBgX(bg, f.visibleArea.xPosition);
        //y: getBgY(f.visibleArea.yPosition);
        y: -26
        height: f.height
        visible: useBackground;
        width: world.width/4;
        transformOrigin: Item.TopLeft;
        scale: Math.min(parent.width/f.width, parent.height/f.height)
        GroundGrassBackground {
            id: ggbg
            x: 0;
            y: bg.height-height
            height: 120
            visible: useBackground;
            width: parent.width;
        }
    }


    function getBgX(bgc, xPos, w)
    {
        var x;
        x=bgc.width*xPos*(bgc.width/world.width);
        if (x>bgc.width-gameArea.width)
            x=bgc.width-gameArea.width;
        return (x<0) ? 0: -x;
    }

    function getBgY(base, yPos)
    {        
        return base+(yPos*32);
    }

    Button {
        size: 64;
        x: 8
        y: 8
        z: 10
        source: "images/buttons/menu.svg";
        visible: gameArea.state!=='paused' && gamePausedDialog.state==='hidden' && gameEnd.state==='hidden'
        onClicked: {
            pauseGame();
        }
    }

    GameTimer {
        id: gameTimer
        anchors.horizontalCenter: parent.horizontalCenter;
        anchors.top: parent.top
        anchors.leftMargin: 4
        anchors.topMargin: 4
        z: 2
        visible: gameScore.visible;
        running: gameArea.state==='playing' && finished===false;
    }

    ScoreDisplay {
        id: gameScore
        anchors.horizontalCenter: parent.horizontalCenter;
        anchors.top: gameTimer.bottom
        anchors.topMargin: 8
        z: 2
        score: player.score;
        visible: gameArea.state==='playing' && displayScore;        
    }

    Flickable {
        id: f
        //anchors.fill: parent;
        contentHeight: world.height;
        contentWidth: world.width;
        interactive: false;
        contentX: getWorldCenterX(cycle ? cycle.playerX : 0)
        contentY: getWorldCenterY(cycle ? cycle.playerY : 0);
        boundsBehavior: Flickable.StopAtBounds
        //clip: true;
        x: 0;
        y: 0;
        width: 800;
        height: 480;
        transformOrigin: Item.TopLeft;
        scale: Math.min(parent.width/width, parent.height/height)

        function getWorldCenterX(cx)
        {
            var x=cx-f.width/5;
            if (x<0)
                x=0;
            if (x>contentWidth-width)
                x=contentWidth-width;

            return x;
        }

        property int bgYmax: 300;
        property int bgYmin: 200;

        function getWorldCenterY(cy)
        {
            var y=bgYmin+cy/5;
            if (y>bgYmax)
                y=bgYmax;
            else if (y<bgYmin)
                y=bgYmin;
            return y;
        }

        function getWorldZoom(speed) {
            var r=Math.abs(1-speed/400);
            if (contentX>contentWidth-width*2)
                r=1;
            return Math.max(0.8, r)
        }

        World {
            id: world;
            width: 23000;
            height: 800;
            running: gameArea.state==='playing';
            velocityIterations: _isFast ? 10 : 8;
            positionIterations: _isFast ? 10 : 8;

            property real gameGravity: -9.81;
            property real gameZoom: 1;            

            gravity {
                x: 0;
                y: gameGravity;
            }

            transform: Scale {
                id: zoom
                origin.x: cycle ? cycle.playerX : 10;
                origin.y: cycle ? cycle.playerY+cycle.height/1.2 : 0;
                xScale: useZoom ? world.gameZoom : 1;
                yScale: xScale;
                Behavior on xScale {
                    enabled: useZoom
                    NumberAnimation { easing.type: Easing.InOutSine; duration: 1250 }
                }                
            }

            function zoomIn()
            {
                gameZoom+=0.5;
            }

            function zoomNormal()
            {
                gameZoom=1.1;
            }

            function zoomNormalSpeed()
            {
                gameZoom=1.0;
            }

            function zoomClose()
            {
                gameZoom=2.5;
            }

            function zoomOut()
            {
                gameZoom-=0.5;
            }

            function zoomOutDebug()
            {
                gameZoom=0.1;
            }

            LevelLoader {
                id: llm;

                onLevelLoaded: {
                    world.loadScene();
                    world.loadItems();
                    world.loadCycle(playerx);
                    gameArea.gameLoaded();
                    gameArea.visible=true;
                }
                onLevelError: {
                    gameArea.gameError();
                }
            }

            function loadScene()
            {                                
                var crate=createSceneItem("Crate.qml");
                var jump=createSceneItem("Jump.qml");
                var plank=createSceneItem("Plank.qml");
                var landing=createSceneItem("Landing.qml");
                var tree=createSceneItem("Tree.qml");
                var cloud=createSceneItem("Cloud.qml");
                var manhole=createSceneItem("ManHole.qml");
                var roadsign=createSceneItem("RoadSign.qml");
                var blocker=createSceneItem("Blocker.qml");

                console.debug("Loading level scene items");

                for (var i=0;i<llm.count;i++) {
                    var go=llm.getItem(i);
                    var y=go.y;

                    if (go.n==='crate') {                        
                        y=ground.y-64-y;
                        crate.createObject(world, {"x": go.x, "y": y, "size": go.s });
                    } else if (go.n==='blocker') {
                        y=ground.y-32-y;
                        blocker.createObject(world, {"x": go.x, "y": y, "size": go.s });
                    } else if (go.n==='jump') {
                        if (y<0)
                            y=ground.y;
                        jump.createObject(world, {"x": go.x, "y": y, "size": go.s, "angle": go.r });
                    } else if (go.n==='landing') {
                        if (y<0)
                            y=ground.y;
                        landing.createObject(world, {"x": go.x, "y": y, "size": go.s, "angle": go.r });
                    } else if (go.n==='plank') {
                        if (y<0)
                            y=ground.y;
                        plank.createObject(world, {"x": go.x, "y": y, "size": go.s });
                    } else if (go.n==='tree1') {
                        if (y<0)
                            y=ground.y-512;
                        tree.createObject(world, {"x": go.x, "y": y, "z": 0, "size": go.s, "type": 1 });
                    } else if (go.n==='tree2') {
                        if (y<0)
                            y=ground.y-384;
                        tree.createObject(world, {"x": go.x, "y": y, "z": 0, "size": go.s, "type": 2 });
                    } else if (go.n==='tree3') {
                        if (y<0)
                            y=ground.y-384;
                        tree.createObject(world, {"x": go.x, "y": y, "z": 0, "size": go.s, "type": 3 });
                    } else if (go.n==='cloud') {
                        y=ground.y-360+go.y;
                        //cloud.createObject(world, {"x": go.x, "y": y, "z": 1, "size": go.s, "animate": world.running, "toX": 0, "fromX": go.x  });
                    } else if (go.n==='manhole') {
                        if (y<0)
                            y=ground.y+10;
                        manhole.createObject(world, {"x": go.x, "y": y, "z": 1 });
                    } else if (go.n==='sign') {
                        roadsign.createObject(world, {"x": go.x, "y": ground.y-128, "type": go.t });
                    }
                }
            }

            function createSceneItem(s)
            {
                var i=Qt.createComponent(s);
                if (i.status !== Component.Ready) {
                    console.debug(s+" failed to load "+i.errorString())
                }
                return i;
            }

            function loadItems()
            {
                console.debug("Creating items to eat");
                B.initFoodItems();
                var gi=Qt.createComponent("GameItem.qml");
                for (var i=0;i<llm.count;i++) {
                    var gid=llm.getItem(i);
                    if (gid.n==='food')
                        gi.createObject(world, {"x": gid.x, "y": ground.y-40-gid.y, "food": B.getFoodItem(gid.t) });
                }
            }

            function loadCycle(startx)
            {
                console.debug("Loading cycle")
                var i=Qt.createComponent("KickScooter.qml");
                if (i.status === Component.Ready) {
                    cycle=i.createObject(world,
                                         {
                                             "x": startx,
                                             "y": ground.y-100,
                                             "z": 5,
                                             "player": player,
                                             "group": -100
                                         });
                    if (cycle) {
                        cycle.itemBad.connect(soundEatBad);
                        cycle.itemGood.connect(soundEatGood);
                        cycle.jumped.connect(soundJump);
                        // cycle.wheelHit.connect(soundWheel);
                        cycle.bodyHit.connect(soundHit);
                        cycle.headHit.connect(soundHeadHit);
                    }
                } else {
                    console.debug("Failed to load bike:"+i.errorString())
                }
            }


            function soundJump(j) { sfx.playSound(j===1 ? 71 : 72); }
            function soundEatGood() { sfx.playSound(62); }
            function soundEatBad() { sfx.playSound(63); }

            property double ghitMinLevel: 72

            function soundWheel(w, fx, fy, f) {
                if (f<ghitMinLevel || Math.abs(fx)<ghitMinLevel) return;
                console.debug("WF:"+f)
                console.debug("WX:"+fx)
                console.debug("WY:"+fy)
                sfx.playSfx(w===1 ? 83 : 84, G.mapValue(f, 60, 400, 0.05, 1.0));
            }
            function soundHit(w, fx, fy, f) {
                sfx.playSfx(w===1 ? 81 : 82, G.mapValue(f, 40, 500, 0.08, 1.0));
            }
            function soundHeadHit(w, fx, fy, f) {
                if (!player.alive)
                    return;
                sfx.playSfx(91, G.mapValue(f, 5, 90, 0.4, 1.0));
            }

            function restartCycle()
            {                
                if (cycle) {
                    cycle.destroy();
                    loadCycle(llm.playerx);
                }
            }

            Component.onCompleted: {
                llm.setLevel(gameArea.level);
            }

            Item {
                id: ground
                height: world.height/8
                anchors {
                    left: parent.left;
                    right: parent.right;
                    bottom: parent.bottom
                }
            }

            Ground {
                id: groundBody
                height: world.height/8
                anchors.fill: ground;
            }            

            Wall {
                id: leftWall
                width: 10
                anchors { right: parent.left; top: parent.top; bottom: parent.bottom }
            }

            Wall {
                id: rightWall
                width: 10
                anchors { left: parent.right; top: parent.top; bottom: parent.bottom }
            }

            EndFlag {
                x: world.width-width;
                y: 0
                width: 200;
                height: world.height;
                groundY: ground.y

                onPlayerHit: {
                    player.cancelNoStuffTrigger();
                    gameArea.endState='complete';
                    finished=true;
                    stopAll();
                    world.resetVelocities();
                    world.zoomNormal();
                    cycle.stepDown();
                    gameStopTimer.start();
                }
            }

            DebugDraw {
                id: debugDraw
                world: world
                anchors.fill: world
                opacity: 0.4
                visible: debug
            }
        }
    }

    // XXX Animate these
    Item {
        anchors.right: parent.right
        anchors.top: parent.top;

        anchors.rightMargin: 4;
        anchors.topMargin: 4;

        width: parent.width/3;        

        visible: gameArea.state==='playing';

        Meter {
            id: waterMeter
            anchors.right: parent.right
            anchors.top: parent.top;
            meterColor: "blue"
            type: "water"
            value: player.water;
            valueMax: player.initialWater;
            height: gameArea.height/8
        }

        Meter {
            id: energyMeter
            anchors.right: parent.right
            anchors.topMargin: 4;
            anchors.top: waterMeter.bottom;
            meterColor: "green"
            type: "energy"
            value: player.energy;
            valueMax: player.initialEnergy;
            height: gameArea.height/8
        }
    }

    GameMessage {
        id: gameStarting
        text: qsTr("Press to start")
        textDetails: levelDescription
        state: gameArea.state==='starting' ? 'shown' : 'hidden'
        onClicked: gameArea.state='playing';
    }

    LevelPauseMenu {
        id: gamePausedDialog
        onUnpause: {
            gameArea.state='playing';
        }
        onQuitGame: {
            stopGame();
        }
        onRestartLevel: {
            gameRestartLevel();
        }
    }

    LevelEndMenu {
        id: gameEnd

        onRestartLevel: {
            gameRestartLevel();
        }
        onQuitGame: {
            stopGame();
        }
        onNextLevel: {
            gameRestartLevel();
        }
        onShown: {

        }
    }

    TouchControls {
        id: controls
        c: gameArea
        visible: _hasTouch && (gameArea.state==='starting' || gameArea.state==='playing')
        anchors.bottom: parent.bottom
        anchors.right: parent.right
        anchors.rightMargin: 16;
        anchors.bottomMargin: 16;
        onClicked: {
            if (gameArea.state==='starting')
                active=true;
        }
    }
    Text {
        id: fps
        visible: world.reportFps
        text: "FPS: "+world.fps
        font.pixelSize: 12
        anchors.bottom: parent.bottom
        anchors.left: parent.left;
    }
    // Disable swipe gesture if game is not paused
    Binding {
        id: swipe
        target: swipeControl
        property: 'locked'
        value: gameArea.state=='playing' ? true : false;
    }
}

