import QtQuick 1.1
import Box2D 1.0
import net.ubc 1.0
import "game.js" as G

Item {
    id: bike
    parent: world;

    property int centerX: (rWheel.x+fWheel.x)/2
    property int centerY: (rWheel.y+fWheel.y+bodyCenter.y)/3

    property int playerX: cyclist.centerX
    property int playerY: cyclist.centerY

    property bool canJump: false;
    property int jumpForce: 1900;
    property bool fixedRotation: false
    property int speed: 1
    property real wheelSize: 32
    property real freeTorque: inAir ? 1 : pedaling ? 40 : 4;
    property real brakeTorque: 44;
    property bool speedBoost: false;
    property real speedBoostRatio: 2;
    property real pedalForce: forward ? -30 : 10;
    property real pedalTorque: speedBoost ? 120 : 60;
    property bool pedaling: false
    property bool breaking: false
    property bool broken: false
    property bool forward: true;
    property int group: -100

    property bool inAir: !(rWheel.isOnGround || fWheel.isOnGround)
    property int hasNudged: 0;
    property variant player: null;

    property bool debug: false;

    // Behavior on freeTorque { NumberAnimation { duration: 150 } }

    signal cyclistDead();
    signal bikeDead();
    signal hitHead();

    signal jumped(int j);
    signal itemGood();
    signal itemBad();
    signal wheelHit(int w, real fx, real fy, real f);
    signal bodyHit(int w, real fx, real fy, real f);
    signal headHit(int w, real fx, real fy, real f);
    signal trickDone();

    onInAirChanged: {
        if (inAir===false)
            hasNudged=0;
    }

    onHasNudgedChanged: console.debug("HN:"+hasNudged)

    Component.onCompleted: {
        console.debug("Scooter is ready");
        console.debug("Player:"+player);
        jumpTimer.interval=400;
        player.gotFood.connect(playerGotFood);
    }

    function playerGotFood(e)
    {
        if (e>0)
            itemGood();
        else
            itemBad();
    }

    property real leanSpeed: 3
    property real directionChangeLimit: 0.9;

    width: fWheel.size * 4;
    height: fWheel.size * 1.8;

    function pedal() {
        if (player && !player.alive)
            return;
        var js=motor.jointSpeed();
        console.debug("ForwardBrakeJS: "+js)
        if (js<directionChangeLimit) {
            forward=true;
            pedaling=true;
            breaking=false;            
        } else {
            breaking=true;
            pedaling=false;
        }
        return pedaling;
    }

    function goBackward() {
        cyclistToSeat.motorSpeed=leanSpeed;
        forward=false;
        if (inAir) {
            hasNudged++;
            rearFork.applyLinearImpulse(Qt.point(-400/hasNudged, 150), rearFork.getWorldPoint(Qt.point(0,0)))
        }
    }

    function goForward() {
        cyclistToSeat.motorSpeed=-leanSpeed;
        forward=true;
        if (inAir) {
            console.debug("ROT: "+rearFork.rotation)
            hasNudged++;
            rearFork.applyLinearImpulse(Qt.point(400/hasNudged, 150), rearFork.getWorldPoint(Qt.point(4.5, 0)))
        }
    }

    function kickPedal() {
        fWheel.applyTorque(-150);
    }

    function reverse() {
        fWheel.applyTorque(350);
        rWheel.applyTorque(350);
        if (inAir) {
            rearFork.applyLinearImpulse(Qt.point(-500, 0), rearFork.getWorldPoint(Qt.point(1.7,0)))
        }
    }

    function stopPedal() {
        pedaling=false;
    }

    function brake() {
        if (player && !player.alive)
            return;

        var js=motor.jointSpeed();
        console.debug("BackwardsBrakeJS: "+js)

        if (js>-directionChangeLimit) {
            forward=false;
            pedaling=true;
            breaking=false;
        } else {
            pedaling=false;
            breaking=true;
        }
        return true;
    }

    function stopBrake() {
        pedaling=false;
        breaking=false;
    }

    function freeBrakeBack() {
        motor.maxMotorTorque=freeTorque;
    }

    function brakeFront() {
        motor.maxMotorTorque=freeTorque*8;
    }

    function freeBrakeFront() {
        motor.maxMotorTorque=freeTorque;
    }

    function jumpWheel() {
        if (broken || (player && player.alive===false))
            return;
        if (rWheel.isOnGround)
            rWheel.applyLinearImpulse(Qt.point(0, -bike.jumpForce), rWheel.getWorldPoint(Qt.point(0.5,0)))

        if (fWheel.isOnGround)
            fWheel.applyLinearImpulse(Qt.point(0, -bike.jumpForce), fWheel.getWorldPoint(Qt.point(0.5,0)))
    }

    function jump() {
        var jp;

        if (broken || (player && player.alive===false))
            return;

        jp=rearFork.getWorldPoint(Qt.point(2,0))
        // Workaround timer interval strangenes bug
        //jumpTimer.interval=200;

        if (rWheel.isOnGround || fWheel.isOnGround) {
            jumpTimer.start();
            jumpGravity();
            rearFork.applyLinearImpulse(Qt.point(0, -bike.jumpForce), jp);
            jumped(2);
            if (rWheel.isOnGround && !fWheel.isOnGround && rWheel.groundItem==300) {
                console.debug("*** Trick Jump!");
                trickDone();
                player.addScore(10);
            }
        } else if (rearFork.isOnGround) {
            jumpTimer.start();
            jumpGravity();
            rearFork.applyLinearImpulse(Qt.point(0, -bike.jumpForce/1.9), jp);
            jumped(1);
        } else {
            console.debug("inAir, no jump");
            return false;
        }
        return true;
    }

    function stopJump()
    {
        jumpTimer.stop();
        normalGravity();
    }

    /*
    Timer {
        id: jumpTimer
        interval: 400;
        onTriggered: {
            cyclistToSeat.motorSpeed=0;
            normalGravity();
        }
    }
    */
    QTimer {
        id: jumpTimer
        interval: 400
        onTimeout: {
            stop();
            cyclistToSeat.motorSpeed=0;
            normalGravity();
        }
    }


    property real normalGravityScale: 3;
    property real jumpGravityScale: -1;

    function normalGravity()
    {
        console.debug("Normal gravity")
        fWheel.gravityScale=normalGravityScale;
        rWheel.gravityScale=normalGravityScale;
        rearFork.gravityScale=normalGravityScale;
        frontFork.gravityScale=normalGravityScale;
        cyclist.gravityScale=normalGravityScale;
    }

    function jumpGravity()
    {
        console.debug("JumpGravity")
        fWheel.gravityScale=jumpGravityScale;
        rWheel.gravityScale=jumpGravityScale;
        rearFork.gravityScale=jumpGravityScale;
        frontFork.gravityScale=jumpGravityScale;
        cyclist.gravityScale=jumpGravityScale;
    }

    function reset() {
        world.resetVelocities();
    }

    function killBikeFront() {
        broken=true;
        backWheelToFork.release();
        playerVehicleBroke();
    }

    function killBikeBack() {
        broken=true;
        motor.release();
        playerVehicleBroke();
    }

    function killBike() {
        broken=true;
        rForkToBody.release();
        fForkToBody.release();
        bmForkToBody.release();
        killBikeFront();
        killBikeBack();
        playerVehicleBroke();
    }

    function playerVehicleBroke()
    {
        player.brakeDown();
    }

    property real forkSize: fWheel.width/6

    property real fwcx: bike.x + bike.width - fWheel.size/2;
    property real fwcy: bike.y + bike.height - fWheel.size/2;

    Wheel {
        id: fWheel
        parent: world
        x: fwcx
        y: fwcy
        size: wheelSize
        group: bike.group

        onBeginContact: {
            var force=G.getTotalForce(fWheel.linearVelocity.x, fWheel.linearVelocity.y);
            wheelHit(1, fWheel.linearVelocity.x, fWheel.linearVelocity.y, force);
            if (y>1000)
                killBikeFront();
        }
    }

    Body {
        id: frontFork
        parent: bike.parent
        fixedRotation: bike.fixedRotation
        sleepingAllowed: true
        x: fwcx-forkSize/2
        y: fwcy-fWheel.width*1.6;
        height: fWheel.width*1.6;
        width: forkSize;

        function checkHit(o)
        {
            var force=G.getTotalForce(frontFork.linearVelocity.x, frontFork.linearVelocity.y);
            if (force>250 && o.gid===300) {
                console.debug("*** ForkForce: "+force)
                killBike();
                cyclist.releasecyclist();
                cyclistDead();
                bodyHit(1, frontFork.linearVelocity.x, frontFork.linearVelocity.y, force);
            }
        }

        fixtures: [
            Box {
                id: fff
                friction: 1.0;
                density: 1.0;
                groupIndex: bike.group
                anchors.fill: parent
                onBeginContact: frontFork.checkHit(other.parent);
            },
            Box {
                id: hff
                friction: 1.0;
                density: 1.0;
                groupIndex: bike.group
                y: fff.y
                x: fff.x-width;
                height: forkSize;
                width: forkSize*2;
                onBeginContact: frontFork.checkHit(other.parent);
            }
        ]
        Image {
            x: fff.x-hff.width-1;
            y: fff.y;
            width: height*0.31;
            height: frontFork.height+2
            source: "images/bike/ohjaustanko.svg"
            z: 100;
        }
    }

    property real rwcx: bike.x + fWheel.size/2;
    property real rwcy: bike.y + bike.height - fWheel.size/2;

    Wheel {
        id: rWheel
        parent: world
        x: rwcx
        y: rwcy
        size: wheelSize
        group: bike.group

        onBeginContact: {
            var force=G.getTotalForce(rWheel.linearVelocity.x, rWheel.linearVelocity.y);            
            wheelHit(2, rWheel.linearVelocity.x, rWheel.linearVelocity.y, force);
            if (y>1000) {
                killBikeBack();
                bikeDead();
            }
        }
    }

    Body {
        id: rearFork
        parent: bike.parent
        fixedRotation: bike.fixedRotation
        sleepingAllowed: true
        x: rwcx
        y: rwcy
        width: fWheel.size*2.3;
        height: forkSize;
        rotation: 0

        property int gcnt: 0
        property bool isOnGround: gcnt>0;

        fixtures: Box {
            id: rff
            friction: 2.0;
            density: 2.0;
            groupIndex: bike.group
            anchors.fill: parent
            onBeginContact: {
                var p=other.parent;
                if (p.gid===500 || p.gid===300) {
                    rearFork.gcnt++;
                }
            }
            onEndContact: {
                var p=other.parent;
                if (p.gid===500 || p.gid===300) {
                    rearFork.gcnt--;
                }
            }
        }
        Image {
            id: runko2
            x: rff.x-height/4-2;
            y: rff.y-height/2.4;
            width: parent.width+4;
            height: (46/256)*width;
            source: "images/bike/runko2.svg"
        }
        Image {
            x: runko2.x+runko2.width*0.98;
            y: runko2.y+runko2.height-height*1.02;
            width: fWheel.size/1.3;
            height: width*(183/118)
            rotation: 0;
            source: "images/bike/runko1.svg"
        }
    }

    Body {
        id: bodyCenter
        parent: bike.parent
        fixedRotation: bike.fixedRotation
        sleepingAllowed: true
        x: rwcx+fWheel.size*2.1
        y: rwcy;
        width: fWheel.width*1.3;
        height: forkSize;
        rotation: -52
        angularDamping: 120.0
        bodyType: Body.Dynamic

        fixtures: Box {
            friction: 6.0;
            density: 10.0;
            anchors.fill: parent

        }
    }

    Body {
        id: bodyMiddleFork
        parent: bike.parent
        fixedRotation: bike.fixedRotation
        sleepingAllowed: true
        x: fwcx-fWheel.size/2
        y: fwcy-fWheel.size/1.5
        width: fWheel.size/2;
        height: forkSize/4
        rotation: 0
        angularDamping: 12.0

        fixtures: Box {
            friction: 10.0;
            density: 2.0;
            anchors.fill: parent

        }
    }

    WeldJoint {
        id: rForkToBody
        parent: bike.parent
        bodyA: rearFork
        bodyB: bodyCenter
        dampingRatio: 1.0
    }

    WeldJoint {
        id: fForkToBody
        parent: bike.parent
        bodyA: frontFork
        bodyB: bodyCenter
        dampingRatio: 1.0
    }

    WeldJoint {
        id: bmForkToBody
        parent: bike.parent
        bodyA: bodyMiddleFork
        bodyB: bodyCenter
        dampingRatio: 1.0
    }

    WheelJoint {
        id: backWheelToFork
        parent: bike.parent
        maxMotorTorque: breaking ? brakeTorque : freeTorque;
        motorSpeed: pedaling && !breaking ? pedalForce : 0;
        enableMotor: true;
        frequencyHz: 60;
        dampingRatio: 1;
        bodyA: rearFork;
        bodyB: rWheel;
    }

    WheelJoint {
        id: motor
        parent: bike.parent
        maxMotorTorque: breaking ? brakeTorque/2 : freeTorque;
        motorSpeed: pedaling && !breaking ? pedalForce : 0;
        enableMotor: true;
        frequencyHz: 60;
        dampingRatio: 1;
        bodyA: frontFork;
        bodyB: fWheel;
    }

    function wheelMotor(pedaling, breaking, onground)
    {
        if (pedaling && !breaking)
            return true;
        return false;
    }

    Cyclist {
        id: cyclist

        x: fwcx - fWheel.size*1.9;
        y: fwcy - fWheel.size*2.5;

        width: fWheel.size*1.1;
        height: fWheel.size*2.3
        parent: world

        player: bike.player;

        group: bike.group

        onBeginContact: {
            if (other.gid===500) {
                var force=G.getTotalForce(x, y);
                if (force>16500) {
                    killCyclist();
                }
            }
        }
        onHitHead: {
            var force=G.getTotalForce(x, y);            
            console.debug("*** HHF: "+force)
            switch (other.gid) {
            case 666:
            case 1000:
            case 2000:
                return;
            case 500:
            case 300: {
                headHit(2, x, y, force);
                if (force>50)
                    killCyclist(force);
            }
            break;
            default: {
                headHit(1, x, y, force);
                if (force>100) {
                    killCyclist(force);
                }
            }
            break;
            }
        }

        function killCyclist(force)
        {
            headStarBurst(force/3);
            makeRagdoll();
            releasecyclist();
            stopPedal();
            stopBrake();
            player.die();
        }

        function releasecyclist()
        {
            hand1ToFork.release();
            hand2ToFork.release();
            cyclistToSeat.release();
        }
    }

    function stepDown()
    {
        //hand2ToFork.enableMotor=false;
        //hand1ToFork.enableMotor=false;
        cyclistToSeat.release();
    }

    /*
    WheelJoint {
        id: hand1ToFork
        parent: bike.parent
        bodyA: frontFork
        bodyB: cyclist.getHand1();
        maxMotorTorque: 20;
        motorSpeed:  0;
        enableMotor: true;
        frequencyHz: 60;
    }

    WheelJoint {
        id: hand2ToFork
        parent: bike.parent
        bodyA: frontFork
        bodyB: cyclist.getHand2();
        maxMotorTorque: 20;
        motorSpeed:  0;
        enableMotor: true;
        frequencyHz: 60;
        dampingRatio: 1;
    }
    */

    WeldJoint {
        id: hand1ToFork
        parent: bike.parent
        bodyA: frontFork
        bodyB: cyclist.getHand1();
    }

    WeldJoint {
        id: hand2ToFork
        parent: bike.parent
        bodyA: frontFork
        bodyB: cyclist.getHand2();
    }

    PrismaticJoint {
        id: cyclistToSeat
        parent: bike.parent
        bodyA: rearFork
        bodyB: cyclist.getFeet1();
        maxMotorForce: 400;
        upperTranslation: 1
        lowerTranslation: -1
        enableMotor: inAir;
        motorSpeed: 0;
        axis: Qt.point(1.0, 0.0)
        enableLimit: true
    }

}
