/*
  Player

  This is the current player object that hold the water and energy levels.
  */
import QtQuick 1.1

Item  {
    id: player
    property real energy: 0;
    property real water: 0;
    property int score: 0;   
    property bool alive: true
    property bool vehicle: true
    property bool playing: false
    property string playerName: ""

    signal noEnergyLeft();
    signal noWaterLeft();
    signal vechicleBroke();
    signal playerDead();

    signal gotFood(int e);

    property bool haveWater: water>0;
    property bool haveEnergy: energy>0;

    property bool criticalWater: water<150;
    property bool criticalEnergy: energy<150;

    property real initialEnergy: 2000;
    property real initialWater: 4000;

    property real pedalEnergyUse: 10
    property real pedalWaterUse: 0.1

    property int dropLevel1: 200;
    property int dropLevel2: 1000;
    property int dropLevel3: 1500;
    property int dropEnergyLevel1: 333;

    property int recycled: 0; // Bottles collected

    Component.onCompleted: {
        reset();
    }

    //onWaterChanged: console.debug("WATER: "+water)
    //onRecycledChanged: console.debug("RECYCLED: "+recycled)
    //onEnergyChanged: console.debug("ENERGY: "+energy)

    // Reset player data for new game
    function reset() {
        energy=initialEnergy;
        water=initialWater;
        score=0;
        recycled=0;
        playing=false;
        alive=true;
        vehicle=true;
    }

    function die()
    {
        if (!alive || !vehicle)
            return;

        alive=false;
        playerDead();
    }

    function brakeDown()
    {
        if (!alive || !vehicle)
            return;

        vehicle=false;
        vechicleBroke();
    }

    // Eat something, consuming water and gaining energy
    // If water level goes to zero, signal it
    // Will only eat if required or if there is enough water to eat with
    // e = energy, w = water, s = score, f = food name
    function eat(e,w,s,f) {
        if (!alive)
            return false;        

        addScore(s);
        gotFood(s);

        // No water at all left, skip
        if (water<=0.0)
            return false;

        if (f.type==='orangejuice') {
            water+=w;
            recycled++;
        } else {
            // Not enough water to eat, skip
            if (water-w<=0.0)
                water=0;
            else
                water-=w;
        }

        // Have enough energy already, skip adding, wasted water, top to max value
        if (energy+e>initialEnergy) {
            energy=initialEnergy;
        } else {
            energy+=e;
        }

        if (water<0)
            water=0;

        // Cancel timer in case we where lucky and got something to eat just in time.
        if (water>0 && energy>0)
            cancelNoStuffTrigger();

        return true;
    }

    function useEnergy(e) {
        if (energy<=0)
            return false;

        energy-=e
        if (energy<0)
            energy=0;        

        return true;
    }

    onHaveEnergyChanged: {
        if (!haveEnergy)
            if (noStuffLeftTrigger.running===false)
                noStuffLeftTrigger.start();
    }

    onHaveWaterChanged: {
        if (!haveWater)
            if (noStuffLeftTrigger.running===false)
                noStuffLeftTrigger.start();
    }

    function useWater(w) {
        water-=w;
    }

    function startUseEnergy()
    {
        playerEnergyTimer.start();
    }

    function stopUseEnergy()
    {
        playerEnergyTimer.stop();
    }

    function getDroplets()
    {
        var drops=0;        

        if (water<dropLevel1)
            return 0;

        if (water>=dropLevel1 && water<=dropLevel2 && recycled>0)
            drops++;
        if (water>dropLevel2 || recycled>4)
            drops++;
        if (water>dropLevel3 || dropEnergyLevel1>333 || recycled>6)
            drops++;
        return drops;
    }

    function getScore()
    {
        return score;
    }

    function getScoreBonus()
    {
        return getScoreEnergy()+getScoreWater()+(recycled*10);
    }

    function getScoreWater()
    {
        if (water<1)
            return 0;
        return Math.floor(water*7)
    }

    function getScoreEnergy()
    {
        if (energy<1)
            return 0;
        return Math.floor(Math.log(energy+1)*657);
    }

    function addScore(s)
    {        
        score+=s;
        if (score<0)
            score=0;
    }

    function cancelNoStuffTrigger()
    {
        noStuffLeftTrigger.stop();
    }

    Timer {
        id: noStuffLeftTrigger
        repeat: false;
        interval: 2000
        onTriggered: {
            console.debug("**************** NoStuffLeftTriggered! ***");
            if (!haveEnergy)
                noEnergyLeft();
            else if (!haveWater)
                noWaterLeft();
        }
    }

    Timer {
        id: playerEnergyTimer
        running: playing
        repeat: true;
        interval: 100
        triggeredOnStart: false;
        onTriggered: {
            useEnergy(pedalEnergyUse);
            useWater(pedalWaterUse);
        }
    }
}
